/*     This program is distributed in the hope that it will be useful,   */
/*     but WITHOUT ANY WARRANTY; without even the implied warranty of    */ 
/*     but WITHOUT ANY WARRANTY; without even the implied warranty of    */
/*     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the      */
/*     GNU General Public License for more details.                      */

/*     You should have received a copy of the GNU General Public License */
/*     along with this program; if not, write to the Free Software       */
/*     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.         */
/*
 * The implementation is derived from the SPOOLES sample described in
 * AllInOne.ps
 *  created -- 98jun04, cca
 *
 * Converted to something that resembles C and
 * support for multithreaded solving added.
 * (C) 2003 Manfred Spraul
 */

#include <stdio.h>
#include <math.h>
@ -21,279 +30,524 @@
#include <misc.h>
#include <FrontMtx.h>
#include <SymbFac.h>
#include <MT/spoolesMT.h>
#include "CalculiX.h"

void spooles(double *ad, double *au, double *b, int *icol, int *irow, 
	     int *neq, int *nzs){
  
  int ipoint,ipo;
  DenseMtx        *mtxB, *mtxX ;
  Chv             *rootchv ;
  ChvManager      *chvmanager  ;
  SubMtxManager   *mtxmanager  ;
  FrontMtx        *frontmtx ;
  InpMtx          *mtxA ;
  double          tau = 100.;
  double          cpus[10] ;
  ETree           *frontETree ;
  FILE            *msgFile ;
  Graph           *graph ;
  int             jrow, jrhs, msglvl=0, ncol, nedges,error, 
                  nent, neqns, nrhs, nrow, pivotingflag=1, seed=7892713, 
                  symmetryflag=0, type=1,row,col,maxdomainsize,maxzeros,
                  maxsize;
  int             *newToOld, *oldToNew ;
  int             stats[20] ;
  IV              *newToOldIV, *oldToNewIV ;
  IVL             *adjIVL, *symbfacIVL ;
int num_cpus = -1;

  /* solving the system of equations using spooles */
/* increase this for debugging */
#define DEBUG_LVL	0

  printf("Solving the system of equations using spooles\n\n");

/*
   --------------------------------------------------
   all-in-one program to solve A X = B
   (1) read in matrix entries and form DInpMtx object
   (2) form Graph object
   (3) order matrix and form front tree
   (4) get the permutation, permute the matrix and 
       front tree and get the symbolic factorization
   (5) compute the numeric factorization
   (6) read in right hand side entries
   (7) compute the solution

   created -- 98jun04, cca
   --------------------------------------------------
*/
    if ( (msgFile = fopen("spooles.out", "a")) == NULL ) {
      fprintf(stderr, "\n fatal error in spooles.c"
	      "\n unable to open file spooles.out\n") ;
    }
/*
   --------------------------------------------
   STEP 1: read the entries from the input file 
           and create the InpMtx object
   --------------------------------------------
*/
    nrow=*neq;
    ncol=*neq;
    nent=*nzs+*neq;
    neqns=nrow;
    ipoint=0;
    mtxA = InpMtx_new() ;
    InpMtx_init(mtxA, INPMTX_BY_ROWS, type, nent, neqns) ;
    for(row=0;row<nrow;row++){
      InpMtx_inputRealEntry(mtxA,row,row,ad[row]);
      for(ipo=ipoint;ipo<ipoint+icol[row];ipo++){
	col=irow[ipo]-1;
	InpMtx_inputRealEntry(mtxA,row,col,au[ipo]);
      }
      ipoint=ipoint+icol[row];
    }
	
    InpMtx_changeStorageMode(mtxA, INPMTX_BY_VECTORS) ;
    if ( msglvl > 1 ) {
      fprintf(msgFile, "\n\n input matrix") ;
      InpMtx_writeForHumanEye(mtxA, msgFile) ;
      fflush(msgFile) ;
    }
/*--------------------------------------------------------------------*/
/*
   -------------------------------------------------
   STEP 2 : find a low-fill ordering
   (1) create the Graph object
   (2) order the graph using multiple minimum degree
   -------------------------------------------------
*/
    graph = Graph_new() ;
    adjIVL = InpMtx_fullAdjacency(mtxA) ;
    nedges = IVL_tsize(adjIVL) ;
    Graph_init2(graph, 0, neqns, 0, nedges, neqns, nedges, adjIVL,
		NULL, NULL) ;
    if ( msglvl > 1 ) {
      fprintf(msgFile, "\n\n graph of the input matrix") ;
      Graph_writeForHumanEye(graph, msgFile) ;
      fflush(msgFile) ;
    }
    /*maxdomainsize=neqns/100;*/
    /* frontETree = orderViaMMD(graph, seed, msglvl, msgFile) ; */
    /* frontETree = orderViaND(graph,maxdomainsize,seed,msglvl,msgFile); */
    /*frontETree = orderViaMS(graph,maxdomainsize,seed,msglvl,msgFile);*/
    maxdomainsize=800;maxzeros=1000;maxsize=64;
    frontETree=orderViaBestOfNDandMS(graph,maxdomainsize,maxzeros,
      maxsize,seed,msglvl,msgFile);
    if ( msglvl > 1 ) {
      fprintf(msgFile, "\n\n front tree from ordering") ;
      ETree_writeForHumanEye(frontETree, msgFile) ;
      fflush(msgFile) ;
#define TUNE_MAXZEROS		1000
#define TUNE_MAXDOMAINSIZE	800
#define TUNE_MAXSIZE		64

#define RNDSEED		7892713
#define MAGIC_DTOL		0.0
#define MAGIC_TAU		100.0

/*
 * Substeps for solving A X = B:
 *
 *  (1) form Graph object
 *  (2) order matrix and form front tree
 *  (3) get the permutation, permute the matrix and 
 *      front tree and get the symbolic factorization
 *  (4) compute the numeric factorization
 *  (5) read in right hand side entries
 *  (6) compute the solution
 *
 * The ssolve_main functions free the input matrices internally
 */

static void ssolve_creategraph(Graph ** graph, ETree ** frontETree,
			 InpMtx * mtxA, int size, FILE * msgFile)
{
	IVL *adjIVL;
	int nedges;

	*graph = Graph_new();
	adjIVL = InpMtx_fullAdjacency(mtxA);
	nedges = IVL_tsize(adjIVL);
	Graph_init2(*graph, 0, size, 0, nedges, size, nedges, adjIVL,
		    NULL, NULL);
	if (DEBUG_LVL > 1) {
		fprintf(msgFile, "\n\n graph of the input matrix");
		Graph_writeForHumanEye(*graph, msgFile);
		fflush(msgFile);
	}
	/* (2) order the graph using multiple minimum degree */

	/*maxdomainsize=neqns/100; */
	/* *frontETree = orderViaMMD(*graph, RNDSEED, DEBUG_LVL, msgFile) ; */
	/* *frontETree = orderViaND(*graph,maxdomainsize,RNDSEED,DEBUG_LVL,msgFile); */
	/* *frontETree = orderViaMS(*graph,maxdomainsize,RNDSEED,DEBUG_LVL,msgFile); */

	*frontETree =
	    orderViaBestOfNDandMS(*graph, TUNE_MAXDOMAINSIZE,
				  TUNE_MAXZEROS, TUNE_MAXSIZE, RNDSEED,
				  DEBUG_LVL, msgFile);
	if (DEBUG_LVL > 1) {
		fprintf(msgFile, "\n\n front tree from ordering");
		ETree_writeForHumanEye(*frontETree, msgFile);
		fflush(msgFile);
	}
}
/*--------------------------------------------------------------------*/
/*
   -----------------------------------------------------
   STEP 3: get the permutation, permute the matrix and 
           front tree and get the symbolic factorization
   -----------------------------------------------------
*/
    oldToNewIV = ETree_oldToNewVtxPerm(frontETree) ;
    oldToNew = IV_entries(oldToNewIV) ;
    newToOldIV = ETree_newToOldVtxPerm(frontETree) ;
    newToOld   = IV_entries(newToOldIV) ;
    ETree_permuteVertices(frontETree, oldToNewIV) ;
    InpMtx_permute(mtxA, oldToNew, oldToNew) ;
    InpMtx_mapToUpperTriangle(mtxA) ;
    InpMtx_changeCoordType(mtxA,INPMTX_BY_CHEVRONS);
    InpMtx_changeStorageMode(mtxA,INPMTX_BY_VECTORS);
    symbfacIVL = SymbFac_initFromInpMtx(frontETree, mtxA) ;
    if ( msglvl > 1 ) {
      fprintf(msgFile, "\n\n old-to-new permutation vector") ;
      IV_writeForHumanEye(oldToNewIV, msgFile) ;
      fprintf(msgFile, "\n\n new-to-old permutation vector") ;
      IV_writeForHumanEye(newToOldIV, msgFile) ;
      fprintf(msgFile, "\n\n front tree after permutation") ;
      ETree_writeForHumanEye(frontETree, msgFile) ;
      fprintf(msgFile, "\n\n input matrix after permutation") ;
      InpMtx_writeForHumanEye(mtxA, msgFile) ;
      fprintf(msgFile, "\n\n symbolic factorization") ;
      IVL_writeForHumanEye(symbfacIVL, msgFile) ;
      fflush(msgFile) ;
    }
/*--------------------------------------------------------------------*/
/*
   ------------------------------------------
   STEP 4: initialize the front matrix object
   ------------------------------------------
*/
    frontmtx = FrontMtx_new() ;
    mtxmanager = SubMtxManager_new() ;
    SubMtxManager_init(mtxmanager, NO_LOCK, 0) ;
    FrontMtx_init(frontmtx, frontETree, symbfacIVL, type, symmetryflag, 
		  FRONTMTX_DENSE_FRONTS, pivotingflag, NO_LOCK, 0, NULL, 
		  mtxmanager, msglvl, msgFile) ;
/*--------------------------------------------------------------------*/
/*
   -----------------------------------------
   STEP 5: compute the numeric factorization
   -----------------------------------------
*/
    chvmanager = ChvManager_new() ;
    ChvManager_init(chvmanager, NO_LOCK, 1) ;
    DVfill(10, cpus, 0.0) ;
    IVfill(20, stats, 0) ;
    rootchv = FrontMtx_factorInpMtx(frontmtx, mtxA, tau, 0.0, chvmanager,
				    &error,cpus, stats, msglvl, msgFile) ;
    ChvManager_free(chvmanager) ;
    if ( msglvl > 1 ) {
      fprintf(msgFile, "\n\n factor matrix") ;
      FrontMtx_writeForHumanEye(frontmtx, msgFile) ;
      fflush(msgFile) ;
    }
    if ( rootchv != NULL ) {
      fprintf(msgFile, "\n\n matrix found to be singular\n") ;
      exit(-1) ;
    }
    if(error>=0){
      fprintf(msgFile,"\n\nerror encountered at front %d",error);
      exit(-1);
    }
/*--------------------------------------------------------------------*/
/*
   --------------------------------------
   STEP 6: post-process the factorization
   --------------------------------------
*/
    FrontMtx_postProcess(frontmtx, msglvl, msgFile) ;
    if ( msglvl > 1 ) {
      fprintf(msgFile, "\n\n factor matrix after post-processing") ;
      FrontMtx_writeForHumanEye(frontmtx, msgFile) ;
      fflush(msgFile) ;
    }
/*--------------------------------------------------------------------*/
/*
   -----------------------------------------
   STEP 7: read the right hand side matrix B
   -----------------------------------------
*/
    nrhs=1;
    mtxB = DenseMtx_new() ;
    DenseMtx_init(mtxB, type, 0, 0, neqns, nrhs, 1, neqns) ;
    DenseMtx_zero(mtxB) ;
    for ( jrow = 0 ; jrow < nrow ; jrow++ ) {
      for ( jrhs = 0 ; jrhs < nrhs ; jrhs++ ) {
	DenseMtx_setRealEntry(mtxB, jrow, jrhs, b[jrow]) ;
      }
    }

    if ( msglvl > 1 ) {
      fprintf(msgFile, "\n\n rhs matrix in original ordering") ;
      DenseMtx_writeForHumanEye(mtxB, msgFile) ;
      fflush(msgFile) ;
    }
/*--------------------------------------------------------------------*/
/*
   ---------------------------------------------------------
   STEP 8: permute the right hand side into the new ordering
   ---------------------------------------------------------
*/
    DenseMtx_permuteRows(mtxB, oldToNewIV) ;
    if ( msglvl > 1 ) {
      fprintf(msgFile, "\n\n right hand side matrix in new ordering") ;
      DenseMtx_writeForHumanEye(mtxB, msgFile) ;
      fflush(msgFile) ;
    }
/*--------------------------------------------------------------------*/
/*
   -------------------------------
   STEP 9: solve the linear system
   -------------------------------
*/
    mtxX = DenseMtx_new() ;
    DenseMtx_init(mtxX, type, 0, 0, neqns, nrhs, 1, neqns) ;
    DenseMtx_zero(mtxX) ;
    FrontMtx_solve(frontmtx, mtxX, mtxB, mtxmanager,cpus, msglvl, msgFile) ;
    if ( msglvl > 1 ) {
      fprintf(msgFile, "\n\n solution matrix in new ordering") ;
      DenseMtx_writeForHumanEye(mtxX, msgFile) ;
      fflush(msgFile) ;
    }
/*--------------------------------------------------------------------*/
/*
   --------------------------------------------------------
   STEP 10: permute the solution into the original ordering
   --------------------------------------------------------
*/
    DenseMtx_permuteRows(mtxX, newToOldIV) ;
    /* *ipb=DenseMtx_entries(mtxX); */
    if ( msglvl > 1 ) {
      fprintf(msgFile, "\n\n solution matrix in original ordering") ;
      DenseMtx_writeForHumanEye(mtxX, msgFile) ;
      fflush(msgFile) ;
    }
/*--------------------------------------------------------------------*/

    for ( jrow = 0 ; jrow < nrow ; jrow++ ) {
      b[jrow]=DenseMtx_entries(mtxX)[jrow];
    }

/*
   -----------
   free memory
   -----------
*/
    FrontMtx_free(frontmtx) ;
    DenseMtx_free(mtxX) ;
    DenseMtx_free(mtxB) ;
    IV_free(newToOldIV) ;
    IV_free(oldToNewIV) ;
    InpMtx_free(mtxA) ;
    ETree_free(frontETree) ;
    IVL_free(symbfacIVL) ;
    SubMtxManager_free(mtxmanager) ;
    Graph_free(graph) ;

/*--------------------------------------------------------------------*/

    fclose(msgFile);
    
  return;
static void ssolve_permuteA(IV ** oldToNewIV, IV ** newToOldIV,
			 IVL ** symbfacIVL, ETree * frontETree,
			 InpMtx * mtxA, FILE * msgFile)
{
	int *newToOld, *oldToNew;

	*oldToNewIV = ETree_oldToNewVtxPerm(frontETree);
	oldToNew = IV_entries(*oldToNewIV);
	*newToOldIV = ETree_newToOldVtxPerm(frontETree);
	newToOld = IV_entries(*newToOldIV);
	ETree_permuteVertices(frontETree, *oldToNewIV);
	InpMtx_permute(mtxA, oldToNew, oldToNew);
	InpMtx_mapToUpperTriangle(mtxA);
	InpMtx_changeCoordType(mtxA, INPMTX_BY_CHEVRONS);
	InpMtx_changeStorageMode(mtxA, INPMTX_BY_VECTORS);
	*symbfacIVL = SymbFac_initFromInpMtx(frontETree, mtxA);
	if (DEBUG_LVL > 1) {
		fprintf(msgFile, "\n\n old-to-new permutation vector");
		IV_writeForHumanEye(*oldToNewIV, msgFile);
		fprintf(msgFile, "\n\n new-to-old permutation vector");
		IV_writeForHumanEye(*newToOldIV, msgFile);
		fprintf(msgFile, "\n\n front tree after permutation");
		ETree_writeForHumanEye(frontETree, msgFile);
		fprintf(msgFile, "\n\n input matrix after permutation");
		InpMtx_writeForHumanEye(mtxA, msgFile);
		fprintf(msgFile, "\n\n symbolic factorization");
		IVL_writeForHumanEye(*symbfacIVL, msgFile);
		fflush(msgFile);
	}
}

static void ssolve_postfactor(FrontMtx *frontmtx, FILE *msgFile)
{
	FrontMtx_postProcess(frontmtx, DEBUG_LVL, msgFile);
	if (DEBUG_LVL > 1) {
		fprintf(msgFile, "\n\n factor matrix after post-processing");
		FrontMtx_writeForHumanEye(frontmtx, msgFile);
		fflush(msgFile);
	}
}

static void ssolve_permuteB(DenseMtx *mtxB, IV *oldToNewIV, FILE* msgFile)
{
	DenseMtx_permuteRows(mtxB, oldToNewIV);
	if (DEBUG_LVL > 1) {
		fprintf(msgFile,
			"\n\n right hand side matrix in new ordering");
		DenseMtx_writeForHumanEye(mtxB, msgFile);
		fflush(msgFile);
	}
}

static void ssolve_permuteout(DenseMtx *mtxX, IV *newToOldIV, FILE *msgFile)
{
	DenseMtx_permuteRows(mtxX, newToOldIV);
	if (DEBUG_LVL > 1) {
		fprintf(msgFile, "\n\n solution matrix in original ordering");
		DenseMtx_writeForHumanEye(mtxX, msgFile);
		fflush(msgFile);
	}
}

static DenseMtx *ssolve_main(InpMtx * mtxA, DenseMtx * mtxB, int size,
			     FILE * msgFile)
{
	Graph *graph;
	ETree *frontETree;
	DenseMtx *mtxX = NULL;
	IV *newToOldIV, *oldToNewIV;
	IVL *symbfacIVL;
	SubMtxManager *mtxmanager;
	FrontMtx *frontmtx;
	Chv *rootchv;
	double cpus[10];

	/*
	 * STEP 1 : find a low-fill ordering
	 * (1) create the Graph object
	 */
	ssolve_creategraph(&graph, &frontETree, mtxA, size, msgFile);

	/*
	 * STEP 2: get the permutation, permute the matrix and 
	 *      front tree and get the symbolic factorization
	 */
	ssolve_permuteA(&oldToNewIV, &newToOldIV, &symbfacIVL, frontETree,
		     mtxA, msgFile);

	/*
	 * STEP 3: initialize the front matrix object
	 */
	{
		frontmtx = FrontMtx_new();
		mtxmanager = SubMtxManager_new();
		SubMtxManager_init(mtxmanager, NO_LOCK, 0);
		FrontMtx_init(frontmtx, frontETree, symbfacIVL, SPOOLES_REAL,
			      SPOOLES_SYMMETRIC, FRONTMTX_DENSE_FRONTS,
			      SPOOLES_PIVOTING, NO_LOCK, 0, NULL,
			      mtxmanager, DEBUG_LVL, msgFile);
	}

	/* 
	 * STEP 4: compute the numeric factorization
	 */
	{
		ChvManager *chvmanager;
		int stats[20];
		int error;

		chvmanager = ChvManager_new();
		ChvManager_init(chvmanager, NO_LOCK, 1);
		DVfill(10, cpus, 0.0);
		IVfill(20, stats, 0);
		rootchv = FrontMtx_factorInpMtx(frontmtx, mtxA, MAGIC_TAU, MAGIC_DTOL,
						chvmanager, &error, cpus,
						stats, DEBUG_LVL, msgFile);
		ChvManager_free(chvmanager);
		if (DEBUG_LVL > 1) {
			fprintf(msgFile, "\n\n factor matrix");
			FrontMtx_writeForHumanEye(frontmtx, msgFile);
			fflush(msgFile);
		}
		if (rootchv != NULL) {
			fprintf(msgFile, "\n\n matrix found to be singular\n");
			exit(-1);
		}
		if (error >= 0) {
			fprintf(msgFile, "\n\nerror encountered at front %d",
				error);
			exit(-1);
		}
	}
	/*
	 * STEP 5: post-process the factorization
	 */
	ssolve_postfactor(frontmtx, msgFile);

	/*
	 * STEP 6: permute the right hand side into the new ordering
	 */
	{
		DenseMtx_permuteRows(mtxB, oldToNewIV);
		if (DEBUG_LVL > 1) {
			fprintf(msgFile,
				"\n\n right hand side matrix in new ordering");
			DenseMtx_writeForHumanEye(mtxB, msgFile);
			fflush(msgFile);
		}
	}
	/*
	 * STEP 7: solve the linear system
	 */
	{
		mtxX = DenseMtx_new();
		DenseMtx_init(mtxX, SPOOLES_REAL, 0, 0, size, 1, 1, size);
		DenseMtx_zero(mtxX);
		FrontMtx_solve(frontmtx, mtxX, mtxB, mtxmanager, cpus,
			       DEBUG_LVL, msgFile);
		if (DEBUG_LVL > 1) {
			fprintf(msgFile, "\n\n solution matrix in new ordering");
			DenseMtx_writeForHumanEye(mtxX, msgFile);
			fflush(msgFile);
		}
	}
	/*
	 * STEP 8:  permute the solution into the original ordering
	 */
	ssolve_permuteout(mtxX, newToOldIV, msgFile);

	/* cleanup: */
	FrontMtx_free(frontmtx);
	IV_free(newToOldIV);
	IV_free(oldToNewIV);
	IVL_free(symbfacIVL);
	SubMtxManager_free(mtxmanager);
	DenseMtx_free(mtxB);
	InpMtx_free(mtxA);
	ETree_free(frontETree);
	Graph_free(graph);
	return mtxX;
}

#ifdef USE_MT 
static DenseMtx *ssolve_main_MT(InpMtx * mtxA, DenseMtx * mtxB, int size,
				FILE * msgFile)
{
	Graph *graph;
	ETree *frontETree;
	DenseMtx *mtxX = NULL;
	IV *newToOldIV, *oldToNewIV, *ownersIV;
	IVL *symbfacIVL;
	SubMtxManager *mtxmanager;
	FrontMtx *frontmtx;
	Chv *rootchv;
	SolveMap *solvemap ;
	double cpus[10];
	int nthread;

	/*
	 * STEP 1 : find a low-fill ordering
	 * (1) create the Graph object
	 */
	ssolve_creategraph(&graph, &frontETree, mtxA, size, msgFile);

	/*
	 * STEP 2: get the permutation, permute the matrix and 
	 *      front tree and get the symbolic factorization
	 */
	ssolve_permuteA(&oldToNewIV, &newToOldIV, &symbfacIVL, frontETree,
		     mtxA, msgFile);

	/*
	 * STEP 3: Prepare distribution to multiple threads/cpus
	 */
	{
		DV *cumopsDV;
		int nfront;

		nfront = ETree_nfront(frontETree);

		nthread = num_cpus;
		if (nthread > nfront)
			nthread = nfront;

		cumopsDV = DV_new();
		DV_init(cumopsDV, nthread, NULL);
		ownersIV = ETree_ddMap(frontETree, SPOOLES_REAL, SPOOLES_SYMMETRIC,
				       cumopsDV, 1. / (2. * nthread));
		if (DEBUG_LVL > 1) {
			fprintf(msgFile,
				"\n\n map from fronts to threads");
			IV_writeForHumanEye(ownersIV, msgFile);
			fprintf(msgFile,
				"\n\n factor operations for each front");
			DV_writeForHumanEye(cumopsDV, msgFile);
			fflush(msgFile);
		} else {
			fprintf(msgFile, "\n\n Using %d threads\n",
				nthread);
		}
		DV_free(cumopsDV);
	}

	/*
	 * STEP 4: initialize the front matrix object
	 */
	{
		frontmtx = FrontMtx_new();
		mtxmanager = SubMtxManager_new();
		SubMtxManager_init(mtxmanager, LOCK_IN_PROCESS, 0);
		FrontMtx_init(frontmtx, frontETree, symbfacIVL, SPOOLES_REAL,
			      SPOOLES_SYMMETRIC, FRONTMTX_DENSE_FRONTS,
			      SPOOLES_PIVOTING, LOCK_IN_PROCESS, 0, NULL,
			      mtxmanager, DEBUG_LVL, msgFile);
	}

	/*
	 * STEP 5: compute the numeric factorization in parallel
	 */
	{
		ChvManager *chvmanager;
		int stats[20];
		int error;

		chvmanager = ChvManager_new();
		ChvManager_init(chvmanager, LOCK_IN_PROCESS, 1);
		DVfill(10, cpus, 0.0);
		IVfill(20, stats, 0);
		rootchv = FrontMtx_MT_factorInpMtx(frontmtx, mtxA, MAGIC_TAU, MAGIC_DTOL,
						   chvmanager, ownersIV, 0,
						   &error, cpus, stats, DEBUG_LVL,
						   msgFile);
		ChvManager_free(chvmanager);
		if (DEBUG_LVL > 1) {
			fprintf(msgFile, "\n\n factor matrix");
			FrontMtx_writeForHumanEye(frontmtx, msgFile);
			fflush(msgFile);
		}
		if (rootchv != NULL) {
			fprintf(msgFile, "\n\n matrix found to be singular\n");
			exit(-1);
		}
		if (error >= 0) {
			fprintf(msgFile, "\n\n fatal error at front %d", error);
			exit(-1);
		}
	}

	/*
	 * STEP 6: post-process the factorization
	 */
	ssolve_postfactor(frontmtx, msgFile);

	/*
	 * STEP 7: permute the right hand side into the new ordering
	 */
	ssolve_permuteB(mtxB, oldToNewIV, msgFile);

	/*
	 * STEP 8: get the solve map object for the parallel solve
	 */
	{
		solvemap = SolveMap_new();
		SolveMap_ddMap(solvemap, SPOOLES_SYMMETRIC,
			       FrontMtx_upperBlockIVL(frontmtx),
			       FrontMtx_lowerBlockIVL(frontmtx), nthread, ownersIV,
			       FrontMtx_frontTree(frontmtx), RNDSEED, DEBUG_LVL,
			       msgFile);
	}

	/*
	 * STEP 9: solve the linear system in parallel
	 */
	{
		mtxX = DenseMtx_new();
		DenseMtx_init(mtxX, SPOOLES_REAL, 0, 0, size, 1, 1, size);
		DenseMtx_zero(mtxX);
		FrontMtx_MT_solve(frontmtx, mtxX, mtxB, mtxmanager, solvemap,
				  cpus, DEBUG_LVL, msgFile);
		if (DEBUG_LVL > 1) {
			fprintf(msgFile, "\n\n solution matrix in new ordering");
			DenseMtx_writeForHumanEye(mtxX, msgFile);
			fflush(msgFile);
		}
	}

	/*
	 * STEP 10: permute the solution into the original ordering
	 */
	ssolve_permuteout(mtxX, newToOldIV, msgFile);

	/* cleanup: */
	FrontMtx_free(frontmtx);
	DenseMtx_free(mtxB);
	IV_free(newToOldIV);
	IV_free(oldToNewIV);
	InpMtx_free(mtxA);
	ETree_free(frontETree);
	IVL_free(symbfacIVL);
	SubMtxManager_free(mtxmanager);
	Graph_free(graph);
	SolveMap_free(solvemap);
	IV_free(ownersIV);

	return mtxX;
}
#endif


/** 
 * spooles: Main interface between Calculix and spooles:
 *
 * Perform 3 operations:
 * 1) load the matrix and the input vector into InpMtx and DenseMtx
 * 2) calculate the solution with ssolve_main{,_MT}
 * 3) convert the resulting DenseMtx back into the Calculix data
 *    format
 */
void spooles(double *ad, double *au, double *b, int *icol, int *irow,
	     int *neq, int *nzs)
{
	FILE *msgFile;
	int size = *neq;
	DenseMtx *mtxB, *mtxX;
	InpMtx *mtxA;

	printf("Solving the system of equations using spooles\n\n");

	if ((msgFile = fopen("spooles.out", "a")) == NULL) {
		fprintf(stderr, "\n fatal error in spooles.c"
			"\n unable to open file spooles.out\n");
	}
	/*
	 * STEP 1: Create the InpMtx object from the Calculix matrix
	 *      representation
	 */
	/* 1a) sparse matrix A */
	{
		int row, ipoint, ipo;
		int nent;

		mtxA = InpMtx_new();
		nent = *nzs + *neq;	/* estimated # of nonzero entries */
		InpMtx_init(mtxA, INPMTX_BY_ROWS, SPOOLES_REAL, nent, size);
		ipoint = 0;
		for (row = 0; row < size; row++) {
			InpMtx_inputRealEntry(mtxA, row, row, ad[row]);
			for (ipo = ipoint; ipo < ipoint + icol[row]; ipo++) {
				int col = irow[ipo] - 1;
				InpMtx_inputRealEntry(mtxA, row, col,
						      au[ipo]);
			}
			ipoint = ipoint + icol[row];
		}
		InpMtx_changeStorageMode(mtxA, INPMTX_BY_VECTORS);

		if (DEBUG_LVL > 1) {
			fprintf(msgFile, "\n\n input matrix");
			InpMtx_writeForHumanEye(mtxA, msgFile);
			fflush(msgFile);
		}
	}
	/* 1b) rhs vector B
	 * Note that there is only one rhs vector, thus
	 * a bit simpler that the AllInOne example
	 */
	{
		int i;
		mtxB = DenseMtx_new();
		DenseMtx_init(mtxB, SPOOLES_REAL, 0, 0, size, 1, 1, size);
		DenseMtx_zero(mtxB);
		for (i = 0; i < size; i++) {
			DenseMtx_setRealEntry(mtxB, i, 0, b[i]);
		}
		if (DEBUG_LVL > 1) {
			fprintf(msgFile, "\n\n rhs matrix in original ordering");
			DenseMtx_writeForHumanEye(mtxB, msgFile);
			fflush(msgFile);
		}
	}

	/* STEP 2: solve it! */
#ifdef USE_MT
	if (num_cpus < 0) {
		char *env;

		num_cpus = 1;
      		env = getenv("CCX_NPROC");
		if (env)
			num_cpus = atoi(env);
		if (num_cpus < 1)
			num_cpus = 1;
	}
	if (num_cpus > 1) {
		/* do not use the multithreaded solver unless
		 * we have multiple threads - avoid the
		 * locking overhead
		 */
		mtxX = ssolve_main_MT(mtxA, mtxB, size, msgFile);
	} else {
		mtxX = ssolve_main(mtxA, mtxB, size, msgFile);
	}
#else
	mtxX = ssolve_main(mtxA, mtxB, size, msgFile);
#endif
	/* STEP 3: convert the result back to Calculix representation */
	{
		int i;
		for (i = 0; i < size; i++) {
			b[i] = DenseMtx_entries(mtxX)[i];
		}
	}
	/* cleanup */
	DenseMtx_free(mtxX);
	fclose(msgFile);
}

